/*
 * Copyright (c) 2008-2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Wed Jan  9 2008
 */

#include <stdio.h>
#include <stdlib.h>
#include <err.h>

#include "viam/viam_acq_driver.h"
#include "dc1394_common.h"


/* --- local data ---------------------------------------------------- */

static int	viam1394_hardware_list(viam_handle_t vh);
static int	viam1394_hardware_attach(viam_handle_t vh,
			viam_camera_t camera);
static void	viam1394_hardware_detach(viam_handle_t vh,
			viam_camera_t camera);
static int	viam1394_camera_configure(viam_handle_t vh,
			viam_camera_t camera);
static int	viam1394_camera_unconfigure(viam_handle_t vh,
			viam_camera_t camera);
static int	viam1394_datatransmit(viam_handle_t vh, viam_camera_t camera,
			viam_onoff_t on);
static int	viam1394_dataflush(viam_handle_t vh, viam_camera_t camera);
static int	viam1394_capture(viam_handle_t vh, viam_camera_t camera);
static char *	viam1394_capturedata(viam_handle_t vh, viam_camera_t camera,
			size_t *bytes, size_t *rowstride);
static struct timeval
		viam1394_capturetimestamp(viam_handle_t vh,
			viam_camera_t camera);
static int	viam1394_capturedone(viam_handle_t vh, viam_camera_t camera);
static int	viam1394_queryhwmodes(viam_handle_t vh, viam_camera_t camera,
			int offset, viam_hwmode_t *supported, size_t n);
static int	viam1394_queryhwfeature(viam_handle_t vh, viam_camera_t camera,
			viam_filtertype_t filter, unsigned int *methods,
			unsigned int *aumodes);
static int	viam1394_sethwfeaturemode(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			viam_filtermethod_t method,
			viam_filterautomode_t aumode);
static int	viam1394_sethwfeaturevalue(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			const viam_filterargs_obj args);

static void	viam1394_camerainfo(dc1394camera_t *camera);

static int	viam1394_sw_geo(dc1394video_mode_t hwgeo, viam_hwmode_t *swgeo);
static int	viam1394_sw_geo_7(viam_handle_t vh,
			const dc1394format7mode_t *hwgeo, viam_hwmode_t *swgeo);
static int	viam1394_sw_fps(dc1394framerate_t hwfps, viam_hwfps_t *swfps);
static int	viam1394_sw_trigger(dc1394trigger_mode_t hwtrig,
			dc1394trigger_polarity_t pol, viam_hwtrigger_t *swtrig);

const struct viam_acqdev_obj viam_acqdev_entry = {
  "dc1394",
  viamgen1394_init,
  viamgen1394_fini,
  viam1394_hardware_list,
  viam1394_hardware_attach,
  viam1394_hardware_detach,
  viam1394_camera_configure,
  viam1394_camera_unconfigure,
  viam1394_datatransmit,
  viam1394_dataflush,
  viam1394_capture,
  viam1394_capturedata,
  viam1394_capturetimestamp,
  viam1394_capturedone,
  viam1394_queryhwmodes,
  viam1394_queryhwfeature,
  viam1394_sethwfeaturemode,
  viam1394_sethwfeaturevalue
};


/* --- viam1394_hardware_list ---------------------------------------- */

/** \brief Print available camera devices on all buses
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

static int
viam1394_hardware_list(viam_handle_t vh)
{
  return viamgen1394_hardware_list(vh, viam1394_camerainfo);
}


/* --- viam1394_hardware_attach -------------------------------------- */

/** \brief Allocate camera ports and handles
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle to attach.
 *
 * \return 0 on success, or an error.
 */

static int
viam1394_hardware_attach(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_camera_t c;
  uint64_t uid;
  char *end;

  /* parse uid into a 64bits integer. XXX for now assume no unit, but
   * unit support should be added at this point with a camera id like
   * dc1394:0xdeadbeef:n with n the unit number. */
  uid = strtoll(viam_camera_uid(camera), &end, 0);
  if (end == viam_camera_uid(camera) || *end != '\0')
    return viam_seterrno(vh, Eviamlib_BAD_ID);

  c = viamgen1394_hardware_attach(vh, viam_camera_name(camera), uid);
  viam_camera_sethwdata(camera, c);

  if (!c) return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);
  return 0;
}


/* --- viam1394_hardware_detach -------------------------------------- */

/** \brief Cleanup dc1394 handle
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 */

static void
viam1394_hardware_detach(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_hardware_detach(vh, viam_camera_hwdata(camera));
  viam_camera_sethwdata(camera, NULL);
}


/* --- viam1394_camera_configure ------------------------------------- */

/** \brief Setup camera with current configuration parameters.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viam1394_camera_configure(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  const viam_hwmode_t *swmode;
  double fps;
  int w, h;
  int s;

  if (!c) return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);
  if (!(swmode = viam_camera_hwmode(vh, camera)))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  if (viam_hwmode_size(vh, swmode, &w, &h, NULL, NULL))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (viamgen1394_fps(swmode->fps, &fps))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  s = viamgen1394_camera_configure(vh, c, swmode->crop, w, h, swmode->format,
                                   fps, swmode->trigger);
  if (s) return s;

  return 0;
}


/* --- viam1394_camera_unconfigure ----------------------------------- */

/** \brief Reset bank configuration
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viam1394_camera_unconfigure(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  int s;

  if (!c) return 0;
  s = viamgen1394_camera_unconfigure(vh, c);
  if (s)
    warnx("couldn't detach capture info from camera `%s'",
      viam_camera_name(camera));
  return s;
}


/* --- viam_datatransmit --------------------------------------------- */

int
viam1394_datatransmit(viam_handle_t vh, viam_camera_t camera, viam_onoff_t on)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_datatransmit(vh, c, on);
}


/* --- viam_dataflush ------------------------------------------------ */

static int
viam1394_dataflush(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_dataflush(vh, c);
}


/* --- viam1394_capture ---------------------------------------------- */

static int
viam1394_capture(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_capture(vh, c);
}


/* --- viam1394_capturedata ------------------------------------------ */

static char *
viam1394_capturedata(viam_handle_t vh, viam_camera_t camera,
		     size_t *bytes, size_t *rowstride)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) { viam_seterrno(vh, Eviamlib_NOT_CONFIGURED); return NULL; }

  return viamgen1394_capturedata(vh, c, bytes, rowstride);
}


/* --- viam1394_capturetimestamp ------------------------------------- */

static struct timeval
viam1394_capturetimestamp(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) {
    viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
    return (struct timeval){0};
  }

  return viamgen1394_capturetimestamp(vh, c);
}


/* --- viam1394_capturedone ------------------------------------------ */

static int
viam1394_capturedone(viam_handle_t vh, viam_camera_t camera)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_capturedone(vh, c);
}


/* --- viam1394_queryhwmodes ----------------------------------------- */

static int
viam1394_queryhwmodes(viam_handle_t vh, viam_camera_t camera, int offset,
		      viam_hwmode_t *supported, size_t n)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  dc1394video_modes_t modes;
  dc1394format7mode_t f7_mode = { .present = 1 };
  dc1394framerates_t framerates;
  dc1394trigger_modes_t *trigmodes;
  dc1394trigger_polarity_t pol;
  viam_hwtrigger_t trig;
  viam_hwmode_t swmode;
  dc1394error_t s;
  dc1394bool_t p;
  int i, j, count;

  if (!c) { viam_seterrno(vh, Eviamlib_NOT_CONFIGURED); return 0; }

  s = dc1394_video_get_supported_modes(c->handle, &modes);
  if (s != DC1394_SUCCESS) { viam_seterrno(vh, Eviamlib_FWBUS); return 0; }

  count = 0;
  f7_mode.color_codings.num = 0;
  for(i=0; i<modes.num;
      f7_mode.color_codings.num>1?f7_mode.color_codings.num--:i++) {
    if (dc1394_is_video_mode_scalable(modes.modes[i])) {
      /* format7: adapt sizes to supported frame sizes */
      if (!f7_mode.color_codings.num) {
	s = dc1394_format7_get_mode_info(c->handle, modes.modes[i], &f7_mode);
	if (s != DC1394_SUCCESS) continue;
      }

      if (viam1394_sw_geo_7(vh, &f7_mode, &swmode)) continue;
      if (f7_mode.color_codings.num < 2) f7_mode.color_codings.num = 0;
    } else {
      /* other formats than format7 */
      if (viam1394_sw_geo(modes.modes[i], &swmode)) continue;
    }

    /* internal trigger is always available */
    swmode.trigger = VIAM_HWTRIGGER_INTERNAL;

    /* get available framerates */
    if (dc1394_is_video_mode_scalable(modes.modes[i])) {
      /* format7 runs without framerate */
      swmode.fps = VIAM_HWFPS_FREE;
      if (count >= offset && (count-offset)<(signed)n)
	supported[count-offset] = swmode;
      count++;
    } else {
      /* other modes do have a framerate */
      s = dc1394_video_get_supported_framerates(
	c->handle, modes.modes[i], &framerates);
      if (s != DC1394_SUCCESS) {
	if (count >= offset && (count-offset)<(signed)n) {
	  supported[count-offset] = swmode;
	  supported[count-offset].fps = VIAM_HWFPS_INVALID;
	}
	count++;
	continue;
      }

      for(j=0; j<framerates.num; j++) {
	if (count >= offset && (count-offset)<(signed)n) {
	  supported[count-offset] = swmode;
	  if (!viam1394_sw_fps(framerates.framerates[j],
			       &supported[count-offset].fps)) {
	    swmode.fps = supported[count-offset].fps;
	  }
	}
	count++;
      }
    }

    /* get external trigger modes */
    if (c->featset.feature[
	  DC1394_FEATURE_TRIGGER-DC1394_FEATURE_MIN].available != DC1394_TRUE)
      continue;

    s = dc1394_external_trigger_has_polarity(c->handle, &p);
    if (s != DC1394_SUCCESS) continue;

    s = dc1394_external_trigger_get_polarity(c->handle, &pol);
    if (s != DC1394_SUCCESS) continue;

    trigmodes = &c->featset.feature[
      DC1394_FEATURE_TRIGGER-DC1394_FEATURE_MIN].trigger_modes;
    for(j=0; j<trigmodes->num; j++) {
      if (p == DC1394_TRUE) {
	if (!viam1394_sw_trigger(trigmodes->modes[j],
				 DC1394_TRIGGER_ACTIVE_LOW, &trig)) {
	  if (count >= offset && (count-offset)<(signed)n) {
	    supported[count-offset] = swmode;
	    supported[count-offset].trigger = trig;
	  }
	  count++;
	}
	if (!viam1394_sw_trigger(trigmodes->modes[j],
				 DC1394_TRIGGER_ACTIVE_HIGH, &trig)) {
	  if (count >= offset && (count-offset)<(signed)n) {
	    supported[count-offset] = swmode;
	    supported[count-offset].trigger = trig;
	  }
	  count++;
	}
      } else {
	if (!viam1394_sw_trigger(trigmodes->modes[j], pol, &trig)) {
	  if (count >= offset && (count-offset)<(signed)n) {
	    supported[count-offset] = swmode;
	    supported[count-offset].trigger = trig;
	  }
	  count++;
	}
      }
    }
  }
  return (count>=offset)?(count-offset>n?n:count-offset):0;
}


/* --- viam1394_queryhwfeature --------------------------------------- */

static int
viam1394_queryhwfeature(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			unsigned int *methods, unsigned int *aumodes)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_queryhwfeature(vh, c, filter, methods, aumodes);
}


/* --- viam1394_sethwfeaturemode ------------------------------------- */

/** \brief Set hardware feature mode for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] method	Hardware mode.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

static int
viam1394_sethwfeaturemode(viam_handle_t vh, viam_camera_t camera,
			  viam_filtertype_t filter, viam_filtermethod_t method,
			  viam_filterautomode_t aumode)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_sethwfeaturemode(vh, c, filter, method, aumode);
}


/* --- viam1394_sethwfeaturevalue ------------------------------------ */

/** \brief Set harware feature value for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] args	Feature value.
 *
 * \return 0 on success.
 */

static int
viam1394_sethwfeaturevalue(viam_handle_t vh, viam_camera_t camera,
			   viam_filtertype_t filter,
			   const viam_filterargs_obj args)
{
  viamgen1394_camera_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_sethwfeaturevalue(vh, c, filter, args);
}


/* --- viam1394_camerainfo ------------------------------------------- */

static void
viam1394_camerainfo(dc1394camera_t *camera)
{
  printf("[node %s:0x%016llx] %s (%s)\n",
	 viam_acqdev_entry.name,
	 camera->guid, camera->model, camera->vendor);
}


/* --- viam1394_geo -------------------------------------------------- */

static const struct {
  viam_hwmode_t sw;
  dc1394video_mode_t hw;
} geomap[] = {
  { { VIAM_HWSZ_160x120, VIAM_HWFMT_YUV444, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_160x120_YUV444 },

  { { VIAM_HWSZ_320x240, VIAM_HWFMT_YUV422_UYVY, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_320x240_YUV422 },

  { { VIAM_HWSZ_640x480, VIAM_HWFMT_YUV411, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_640x480_YUV411 },

  { { VIAM_HWSZ_640x480, VIAM_HWFMT_YUV422_UYVY, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_640x480_YUV422 },

  { { VIAM_HWSZ_640x480, VIAM_HWFMT_RGB888, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_640x480_RGB8 },

  { { VIAM_HWSZ_640x480, VIAM_HWFMT_MONO8,  VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_640x480_MONO8 },

  { { VIAM_HWSZ_640x480, VIAM_HWFMT_MONO16, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_640x480_MONO16 },

  { { VIAM_HWSZ_800x600, VIAM_HWFMT_YUV422_UYVY, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_800x600_YUV422 },

  { { VIAM_HWSZ_800x600,   VIAM_HWFMT_RGB888, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_800x600_RGB8 },

  { { VIAM_HWSZ_800x600,   VIAM_HWFMT_MONO8,  VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_800x600_MONO8 },

  { { VIAM_HWSZ_1024x768,  VIAM_HWFMT_YUV422_UYVY, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1024x768_YUV422 },

  { { VIAM_HWSZ_1024x768,  VIAM_HWFMT_RGB888, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1024x768_RGB8 },

  { { VIAM_HWSZ_1024x768,  VIAM_HWFMT_MONO8,  VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1024x768_MONO8 },

  { { VIAM_HWSZ_800x600,   VIAM_HWFMT_MONO16, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_800x600_MONO16 },

  { { VIAM_HWSZ_1024x768,  VIAM_HWFMT_MONO16, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1024x768_MONO16 },

  { { VIAM_HWSZ_1280x960,  VIAM_HWFMT_YUV422_UYVY, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1280x960_YUV422 },

  { { VIAM_HWSZ_1280x960,  VIAM_HWFMT_RGB888, VIAM_HW_FIXED } ,
    DC1394_VIDEO_MODE_1280x960_RGB8 },

  { { VIAM_HWSZ_1280x960,  VIAM_HWFMT_MONO8,  VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1280x960_MONO8 },

  { { VIAM_HWSZ_1600x1200, VIAM_HWFMT_YUV422_UYVY, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1600x1200_YUV422 },

  { { VIAM_HWSZ_1600x1200, VIAM_HWFMT_RGB888, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1600x1200_RGB8 },

  { { VIAM_HWSZ_1600x1200, VIAM_HWFMT_MONO8,  VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1600x1200_MONO8 },

  { { VIAM_HWSZ_1280x960,  VIAM_HWFMT_MONO16, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1280x960_MONO16 },

  { { VIAM_HWSZ_1600x1200, VIAM_HWFMT_MONO16, VIAM_HW_FIXED },
    DC1394_VIDEO_MODE_1600x1200_MONO16 }
};

static int
viam1394_sw_geo(dc1394video_mode_t hwgeo, viam_hwmode_t *swgeo)
{
  int i;

  for(i=0; i<sizeof(geomap)/sizeof(geomap[0]); i++)
    if (geomap[i].hw == hwgeo) {
      swgeo->size = geomap[i].sw.size;
      swgeo->format = geomap[i].sw.format;
      swgeo->crop = geomap[i].sw.crop;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}

static int
viam1394_sw_geo_7(viam_handle_t vh, const dc1394format7mode_t *hwgeo,
		  viam_hwmode_t *swgeo)
{
  int i, w, h;
  int maxw = 0, maxh = 0;

  /* find the largest standard size smaller than the max hardware size */
  swgeo->size = VIAM_HWSZ_INVALID;
  for(i=0; i<sizeof(geomap)/sizeof(geomap[0]); i++) {
    if (viam_hwmode_size(vh, &geomap[i].sw, &w, &h, NULL, NULL)) continue;
    if (w > hwgeo->max_size_x && h > hwgeo->max_size_y) continue;

    if (w*h > maxw*maxh) {
      maxw = w;
      maxh = h;
      swgeo->size = geomap[i].sw.size;
    }
  }

  if (swgeo->size == VIAM_HWSZ_INVALID) return Eviamlib_UNSUPPORTED_HWMODE;

  /* map image format */
  switch(hwgeo->color_codings.codings[hwgeo->color_codings.num-1]) {
    case DC1394_COLOR_CODING_MONO8:
      swgeo->format = VIAM_HWFMT_MONO8;		break;
    case DC1394_COLOR_CODING_YUV411:
      swgeo->format = VIAM_HWFMT_YUV411;	break;
    case DC1394_COLOR_CODING_YUV422:
      swgeo->format = VIAM_HWFMT_YUV422_UYVY;	break;
    case DC1394_COLOR_CODING_YUV444:
      swgeo->format = VIAM_HWFMT_YUV444;	break;
    case DC1394_COLOR_CODING_RGB8:
      swgeo->format = VIAM_HWFMT_RGB888;	break;
    case DC1394_COLOR_CODING_MONO16:
      swgeo->format = VIAM_HWFMT_MONO16;	break;
    case DC1394_COLOR_CODING_RAW8:
      swgeo->format = VIAM_HWFMT_MONO8;		break;
    case DC1394_COLOR_CODING_RAW16:
      swgeo->format = VIAM_HWFMT_MONO16;	break;

    default: return Eviamlib_UNSUPPORTED_HWMODE;
  }

  swgeo->crop = VIAM_HW_CROP;
  return 0;
}


/* --- viam1394_fps -------------------------------------------------- */

static const struct {
  viam_hwfps_t sw;
  dc1394framerate_t hw;
  double v;
} fpsmap[] = {
  { VIAM_HWFPS_1_875,	DC1394_FRAMERATE_1_875,	1.875},
  { VIAM_HWFPS_3_75,	DC1394_FRAMERATE_3_75,	3.75},
  { VIAM_HWFPS_7_5,	DC1394_FRAMERATE_7_5,	7.5},
  { VIAM_HWFPS_15,	DC1394_FRAMERATE_15,	15.},
  { VIAM_HWFPS_30,	DC1394_FRAMERATE_30,	30.},
  { VIAM_HWFPS_60,	DC1394_FRAMERATE_60,	60.},
  { VIAM_HWFPS_120,	DC1394_FRAMERATE_120,	120.},
  { VIAM_HWFPS_240,	DC1394_FRAMERATE_240,	240.},
};

static int
viam1394_sw_fps(dc1394framerate_t hwfps, viam_hwfps_t *swfps)
{
  int i;

  for(i=0; i<sizeof(fpsmap)/sizeof(fpsmap[0]); i++)
    if (fpsmap[i].hw == hwfps) {
      *swfps = fpsmap[i].sw;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}


/* --- viam1394_trig ------------------------------------------------- */

static const struct {
  viam_hwtrigger_t sw;
  dc1394trigger_mode_t hw;
  dc1394trigger_polarity_t pol;
} triggermap[] = {
  { VIAM_HWTRIGGER_MODE0_LOW,	DC1394_TRIGGER_MODE_0,	DC1394_TRIGGER_ACTIVE_LOW },
  { VIAM_HWTRIGGER_MODE0_HIGH,	DC1394_TRIGGER_MODE_0,	DC1394_TRIGGER_ACTIVE_HIGH },
  { VIAM_HWTRIGGER_MODE1_LOW,	DC1394_TRIGGER_MODE_1,	DC1394_TRIGGER_ACTIVE_LOW },
  { VIAM_HWTRIGGER_MODE1_HIGH,	DC1394_TRIGGER_MODE_1,	DC1394_TRIGGER_ACTIVE_HIGH }
};

static int
viam1394_sw_trigger(dc1394trigger_mode_t hwtrig, dc1394trigger_polarity_t pol,
		    viam_hwtrigger_t *swtrig)
{
  int i;

  for(i=0; i<sizeof(triggermap)/sizeof(triggermap[0]); i++)
    if (triggermap[i].hw == hwtrig && (triggermap[i].pol&1) == (pol&1)) {
      *swtrig = triggermap[i].sw;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}
